<?php

/*
 * This file is part of the Symfony package.
 *
 * (c) Fabien Potencier <fabien@symfony.com>
 *
 * For the full copyright and license information, please view the LICENSE
 * file that was distributed with this source code.
 */

#[Deprecated(message: 'use UUID_TYPE_RANDOM instead')]
const UUID_TYPE_DCE = 4;

#[Deprecated(message: 'use UUID_TYPE_TIME instead')]
const UUID_TYPE_NAME = 1;
